CHANGE_NOTE = [
    {
        "http://purl.org/dc/terms/creator": [{"@id": "http://example.com/john.doe"}],
        "http://purl.org/dc/terms/issued": [
            {"@type": "http://www.w3.org/2001/XMLSchema#date", "@value": "1999-01-23"}
        ],
        "http://www.w3.org/1999/02/22-rdf-syntax-ns#value": [
            {"@language": "en", "@value": "Moved from under 'fruits' to under 'vegetables'"}
        ],
    },
]
HISTORY_NOTE = [
    {
        "http://purl.org/dc/terms/creator": [{"@id": "http://example.com/jane.doe"}],
        "http://purl.org/dc/terms/issued": [
            {"@type": "http://www.w3.org/2001/XMLSchema#date", "@value": "1971-01-23"}
        ],
        "http://www.w3.org/1999/02/22-rdf-syntax-ns#value": [
            {"@language": "en", "@value": "Previously thought we were talking about horseradish"}
        ],
    },
]
EDITORIAL_NOTE = [
    {
        "http://purl.org/dc/terms/creator": [{"@id": "http://example.com/jane.doe"}],
        "http://purl.org/dc/terms/issued": [
            {"@type": "http://www.w3.org/2001/XMLSchema#date", "@value": "2012-01-23"}
        ],
        "http://www.w3.org/1999/02/22-rdf-syntax-ns#value": [
            {"@language": "en", "@value": "Check if this includes species outside Animalia"}
        ],
    },
]
DEFINITION = [
    {
        "@language": "en",
        "@value": "Includes all sorts of live animals, including things you probably never heard of before, like psittaciformes. Those are birds like parrots, by the way. Also includes silly names like hinnies and asses, but not those kinds of asses, get your mind out of the gutter. We are very serious people doing very serious work with our asses.",
    }
]
