from fastapi import APIRouter, Request
from fastapi.responses import RedirectResponse

router = APIRouter(include_in_schema=False)


@router.get('/')
async def redirect_main_page(
    request: Request,
) -> RedirectResponse:
    return RedirectResponse(request.url_for('web_concept_schemes'))
